/* Base styles */
.graph-container {
    flex: 1;
    min-width: 300px;
    height: 400px;
}

/* Media queries for different zoom levels */
@media screen and (min-width: 1200px) {
    .graph-container {
        min-width: 400px;
        height: 500px;
    }
}

@media screen and (min-width: 1500px) {
    .graph-container {
        min-width: 500px;
        height: 600px;
    }
}

@media screen and (min-width: 1800px) {
    .graph-container {
        min-width: 600px;
        height: 700px;
    }
}